import pytest

from pactum import verbs
from pactum.request import Request


def test_basic_request():
    request = Request(verb=verbs.GET, payload=None, headers=[])

    assert request.verb == verbs.GET
    assert request.payload is None
    assert request.headers == []


def test_basic_request_class_def():
    class TestRequest(Request):
        payload = None
        verb = verbs.GET
        headers = []

    request = TestRequest()

    assert request.verb == verbs.GET
    assert request.payload is None
    assert request.headers == []


def test_fails_if_no_verb():
    with pytest.raises(TypeError):
        Request(payload=None, headers=[])



import pytest

from pactum import verbs
from pactum.action import Action
from pactum.api import API
from pactum.resources import Resource
from pactum.request import Request
from pactum.response import Response
from pactum.route import Route
from pactum.version import Version


@pytest.fixture
def route(resource):
    return Route(path='/test/')


@pytest.fixture
def version(route):
    return Version(
        name="v0",
        routes=[
            route,
        ],
    )


@pytest.fixture
def api():
    api = API(
        name="Test API",
        versions=[],
    )
    return api


@pytest.fixture
def resource():
    return Resource()


@pytest.fixture
def request():
    return Request(verb=verbs.GET)


@pytest.fixture
def response():
    return Response(status=200, body='')


@pytest.fixture
def action(request, response):
    return Action(request=request, responses=[response])



import pytest

from pactum.fields import Field, IntegerField, PositiveIntegerField, ResourceField, StringField, DecimalField
from pactum.resources import Resource


def test_basic_field():
    field = Field(
        name="TestField",
        type="type",
    )

    assert field.name == "TestField"
    assert field.type == "type"


def test_basic_field_class_definition():
    class TestField(Field):
        name = "TestField"
        type = "type"

    field = TestField()
    assert field.name == "TestField"
    assert field.type == "type"


def test_basic_field_class_definition_with_class_name():
    class MyField(Field):
        name = "name"

    field = MyField()
    assert field.name == "name"
    assert field.type is MyField


def test_field_type_can_be_class_name():
    field = Field(
        name="TestField",
    )

    assert field.name == "TestField"
    assert field.type == Field


def test_integer_field():
    int_field = IntegerField(name="my_name")

    assert int_field.type == IntegerField
    assert int_field.name == "my_name"
    assert int_field.min_value is None
    assert int_field.max_value is None


def test_integer_field_class_field():
    class MyIntegerField(IntegerField):
        min_value = -5
        max_value = 5

    int_field = MyIntegerField()
    assert int_field.min_value == -5
    assert int_field.max_value == 5


def test_integer_field_min_max_values():
    int_field = IntegerField(name="my_name", min_value=-5, max_value=5)

    assert int_field.type == IntegerField
    assert int_field.min_value == -5
    assert int_field.max_value == 5


def test_positive_integer_field():
    int_field = PositiveIntegerField(name="my_name")

    assert int_field.type == PositiveIntegerField
    assert int_field.min_value == 0


def test_string_field():
    str_field = StringField(name="my_name")

    assert str_field.type == StringField
    assert str_field.name == "my_name"


def test_resource_field(resource):
    resource_field = ResourceField(name="my_field", resource=resource)

    assert resource_field.type == ResourceField
    assert resource_field.name == "my_field"
    assert resource_field.resource == resource


def test_resource_field_class_def(resource):
    class TestResourceField(ResourceField):
        name = "my_field"
        resource = Resource()

    resource_field = TestResourceField()

    assert resource_field.type == TestResourceField
    assert resource_field.name == "my_field"
    assert isinstance(resource_field.resource, Resource)


def test_fails__forresource_field_missing_ressource(resource):
    class TestResourceField(ResourceField):
        name = "my_field"

    with pytest.raises(TypeError):
        TestResourceField()


def test_basic_decimal_field():
    dec_field = DecimalField(name="dec_field", precision=2)

    assert dec_field.type == DecimalField
    assert dec_field.name == "dec_field"
    assert dec_field.precision == 2


def test_basic_decimal_field_class_definition():
    class MyDecimalField(DecimalField):
        name = "dec_field"
        precision = 2

    dec_field = MyDecimalField()

    assert dec_field.type == MyDecimalField
    assert dec_field.name == "dec_field"
    assert dec_field.precision == 2


def test_fail_basic_decimal_field_no_precision():
    with pytest.raises(TypeError):
        DecimalField(name="dec_field")



import pytest

from pactum.action import Action


def test_basic_action(request, response):
    test_action = Action(request=request, responses=[response], description="Test action")

    assert test_action.request == request
    assert test_action.responses == [response]
    assert test_action.responses[0].parent == test_action
    assert test_action.request.parent == test_action
    assert test_action.__doc__ == "Test action"


def test_basic_action_class_def(response):
    class TestAction(Action):
        request = None
        responses = [response]
        description = "Test Action"

    test_action = TestAction()
    assert test_action.request is None
    assert len(test_action.responses) == 1
    assert test_action.responses[0].parent == test_action
    assert test_action.__doc__ == "Test Action"


def test_basic_action_class_def_with_doc_description(response):
    class TestAction(Action):
        '''Test Action.'''
        request = None
        responses = [response]

    test_action = TestAction()
    assert test_action.request is None
    assert len(test_action.responses) == 1
    assert test_action.responses[0].parent == test_action
    assert test_action.__doc__ == "Test Action."


def test_action_must_have_responses():
    class TestAction(Action):
        request = None

    with pytest.raises(TypeError):
        TestAction()



from .base import Element


class Response(Element):
    def __init__(self, status=None, body=None, headers=None, **kwargs):
        super().__init__(**kwargs)

        if body is None:
            body = getattr(self, 'body', None)
        self.body = body

        if headers is None:
            headers = getattr(self, 'headers', [])
        self.headers = headers

        if status is None:
            try:
                status = getattr(self, 'status')
            except AttributeError:
                raise TypeError('Missing status specification.')
        self.status = status

    def accept(self, visitor):
        visitor.visit_response(self)
        if self.body is not None:
            self.body.accept(visitor)



import importlib
import json

import click
import yaml
from pactum.exporters.openapi import OpenAPIV3Exporter


def clean_dict(d):
    if not isinstance(d, (dict, list)):
        return d
    if isinstance(d, list):
        return [v for v in (clean_dict(v) for v in d) if v]
    return {k: v for k, v in ((k, clean_dict(v)) for k, v in d.items()) if v not in (None, '', [], {})}


@click.command()
@click.argument('spec_file', type=click.Path())
@click.argument('output_file', type=click.File(mode='w'))
@click.option('--format', type=click.Choice(['json', 'yaml']), default='json')
def openapi_export(spec_file, output_file, format):
    spec = importlib.machinery.SourceFileLoader('spec', spec_file).load_module()
    api = spec.api
    visitor = OpenAPIV3Exporter()
    api.accept(visitor)
    result = clean_dict(visitor.result)
    if format == 'json':
        json.dump(result, output_file, indent=2)
    elif format == 'yaml':
        yaml.dump(result, output_file, default_flow_style=False)



from .base import Element


class Route(Element):
    _children_name = 'actions'

    def __init__(self, path=None, actions=None, **kwargs):
        super().__init__(**kwargs)
        if path is None:
            try:
                path = getattr(self, 'path')
            except AttributeError:
                raise TypeError("Missing path specification.")
        self.path = path

        self._initialize_children(locals())

    def accept(self, visitor):
        visitor.visit_route(self)
        for action in self.actions:
            action.accept(visitor)



from .base import Element


class Action(Element):
    _children_name = 'responses'

    def __init__(self, request=None, responses=None, **kwargs):
        super().__init__(**kwargs)

        if request is None:
            request = getattr(self.__class__, 'request', None)
        self.request = request
        if self.request:
            self.request.parent = self

        self.responses = []
        if responses is None:
            try:
                responses = getattr(self.__class__, 'responses')
            except AttributeError:
                raise TypeError('Missing responses specification.')

        self._initialize_children(locals())

    def accept(self, visitor):
        visitor.visit_action(self)
        self.request.accept(visitor)
        for response in self.responses:
            response.accept(visitor)



import pytest
from pactum.response import Response


def test_basic_response():
    response = Response(status=200, body=None, headers=[])

    assert response.body is None
    assert response.headers == []
    assert response.status == 200


def test_basic_response_class_def():
    class TestResponse(Response):
        body = None
        headers = []
        status = 200

    response = TestResponse()

    assert response.body is None
    assert response.headers == []
    assert response.status == 200


def test_response_is_invalid_without_status():
    class TestResponse(Response):
        body = None
        headers = []

    with pytest.raises(TypeError):
        TestResponse()


def test_response_class_definition_with_doc_description(version):
    class TestResponse(Response):
        """Response for tests."""
        body = None
        headers = []
        status = 200

    response = TestResponse()

    assert response.body is None
    assert response.headers == []
    assert response.status == 200
    assert response.__doc__ == "Response for tests."



class Request:
    def __init__(self, verb=None, payload=None, headers=None):
        if verb is None:
            try:
                verb = getattr(self, 'verb')
            except AttributeError:
                raise TypeError('Missing request verb specification.')
        self.verb = verb

        if payload is None:
            payload = getattr(self, 'payload', None)
        self.payload = payload

        if headers is None:
            headers = getattr(self, 'headers', [])
        self.headers = headers

    def accept(self, visitor):
        visitor.visit_request(self)
        if self.payload is not None:
            self.payload.accept(visitor)



from pactum.api import API


def test_base_api():
    api = API(
        name="Test API",
        versions=[],
        description='API for tests.'
    )

    assert api.name == "Test API"
    assert api.__doc__ == "API for tests."
    assert len(api.versions) == 0


def test_base_api_class_definition(version):
    class TestAPI(API):
        name = "Test API"
        versions = [version]
        description = "API for tests."

    api = TestAPI()

    assert api.name == "Test API"
    assert api.__doc__ == "API for tests."
    assert len(api.versions) == 1
    assert api.versions[0].parent == api


def test_api_class_definition_with_doc_description(version):
    class TestAPI(API):
        """
        API for tests.
        """
        name = "Test API"
        versions = [version]

    api = TestAPI()

    assert api.name == "Test API"
    assert api.__doc__ == "API for tests."
    assert len(api.versions) == 1
    assert api.versions[0].parent == api


def test_api_with_one_version(version):
    api = API(
        name="Test API",
        versions=[
            version,
        ],
    )

    assert len(api.versions) == 1
    assert api.versions[0].parent == api


def test_api_class_definition_with_one_version(version):
    class TestAPI(API):
        name = "Test API"
        versions = [
            version,
        ]

    api = TestAPI()
    assert len(api.versions) == 1
    assert api.versions[0].parent == api


def test_api_add_version(api, version):
    api.append(version)
    assert len(api.versions) == 1
    assert api.versions[0].parent == api


def test_prefer_parameter_to_class_definition(version):
    class TestAPI(API):
        name = "Test API"
        versions = [version, version]

    api = TestAPI(name="Test API by parameter", versions=[version])

    assert len(api.versions) == 1
    assert api.name == "Test API by parameter"
    assert api.versions[0].parent == api



from pactum import fields
from pactum.visitor import BaseVisitor

FIELD_TYPES_MAP = {
    fields.IntegerField: 'integer',
    fields.PositiveIntegerField: 'integer',
    fields.StringField: 'string',
    fields.TimestampField: 'datetime',
}


class NotSpecified(Exception):
    pass


class OpenAPIV3Exporter(BaseVisitor):
    '''
    Exporter based on OpenAPI specification version 3.0.1.
    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md
    '''
    OPENAPI_VERSION = '3.0.1'

    def __init__(self):
        self.result = {
            'openapi': self.__class__.OPENAPI_VERSION,  # REQUIRED
            'servers': [],
            'info': {},  # REQUIRED
            'paths': {},  # REQUIRED
            'components': {'schemas': {}},
            'security': {},
            'tags': {}
        }

    def visit_api(self, api):
        last_version = api.versions[-1]
        self.result['info'] = {
            'title': api.name,  # REQUIRED
            'description': api.__doc__,
            'termsOfService': '',
            'contact': {},
            'license': {},
            'version': last_version.name,  # REQUIRED
        }
        api.versions = [last_version]  # Versions override to visitonly children for last version.

    def visit_version(self, version):
        pass

    def visit_route(self, route):
        self.result['paths'][route.path] = {
            'summary': '',
            'description': route.__doc__,
            'servers': [],
            'parameters': {},
            # get, put, post, delete, options, head, patch and trace will be populated on children visits.
        }

    def visit_action(self, action):
        self.result['paths'][action.parent.path][action.request.verb.lower()] = {
            'description': action.__doc__,
            'tags': [],
            'summary': action.__doc__,
            'externalDocs': [],
            'operationId': ''.join(x.title() for x in action.__doc__.split()),
            'parameters': [],
            'responses': {},  # will be populated on children visits.
            'callbacks': [],
            'deprecated': False,
            'security': {},
            'servers': {},
        }

    def visit_request(self, request):
        payload = {}
        if request.payload:
            payload = {'schema': {'$ref': f'#/components/schemas/{request.payload.name}'}}
        self.result['paths'][request.parent.parent.path][request.verb.lower()]['requestBody'] = payload

    def visit_response(self, response):
        method = self.result['paths'][response.parent.parent.path][response.parent.request.verb.lower()]
        parsed_response = {
            'description': response.__doc__,
            'links': {},
            'content': {}
        }

        headers = dict(response.headers)
        content_type = headers.pop('content-type', None)
        parsed_response['headers'] = headers

        if content_type is None:
            content_type = '*/*'

        if response.body:
            parsed_response['content'][content_type] = {'schema': {'$ref': f'#/components/schemas/{response.body.name}'}}
        method['responses'][str(response.status)] = parsed_response

    def visit_resource(self, resource):
        self.result['components']['schemas'][resource.name] = {
            'type': 'object',
            'required': [field.name for field in resource.fields if field.required],
            'properties': {}
        }

    def visit_list_resource(self, list_resource):
        self.result['components']['schemas'][list_resource.name] = {
            'type': 'array', 'items': {'$ref': f'#/components/schemas/{list_resource.resource.name}'}
        }

    def _map_field_type(self, field):
        type_ = FIELD_TYPES_MAP.get(field.type, field.extensions.get('openapi.type'))
        if type_ is None:
            raise NotSpecified(f'Type for field {field.type} is not specified for OpenAPI.')
        return type_

    def visit_field(self, field):
        obj = self.result['components']['schemas'][field.parent.name]
        if field.type == fields.ResourceField:
            obj['properties'][field.name] = {'$ref': f'#/components/schemas/{field.resource.name}'}
        else:
            type_ = self._map_field_type(field)

            obj['properties'][field.name] = {"type": type_}

            if getattr(field, 'min_value', None) is not None:
                obj['properties'][field.name]['minimum'] = field.min_value

            if getattr(field, 'max_value', None) is not None:
                obj['properties'][field.name]['maximum'] = field.max_value



class Element:
    def __init__(self, description=None, extensions=None, **kwargs):
        if description is None:
            description = getattr(self, 'description', self.__doc__ or '')
        self.__doc__ = description.strip()

        if extensions is None:
            extensions = getattr(self, 'extensions', {})
        self.extensions = extensions

    def _initialize_children(self, attrs):
        setattr(self, self._children_name, [])

        children = attrs.get(self._children_name)
        if children is None:
            children = getattr(self.__class__, self._children_name, [])

        for child in children:
            self.append(child)

    def append(self, child):
        child_attr = getattr(self, self._children_name, [])
        child_attr.append(child)
        child.parent = self



class BaseVisitor:
    def visit_api(self, api):
        raise NotImplementedError('visit_api is not implemented')

    def visit_version(self, version):
        raise NotImplementedError('visit_version is not implemented')

    def visit_route(self, route):
        raise NotImplementedError('visit_route is not implemented')

    def visit_action(self, action):
        raise NotImplementedError('visit_action is not implemented')

    def visit_request(self, request):
        raise NotImplementedError('visit_request is not implemented')

    def visit_response(self, response):
        raise NotImplementedError('visit_response is not implemented')

    def visit_resource(self, resource):
        raise NotImplementedError('visit_resource is not implemented')

    def visit_field(self, field):
        raise NotImplementedError('visit_field is not implemented')

    def visit_list_resource(self, list_resource):
        raise NotImplementedError('visit_list_resource is not implemented')



from .base import Element


class API(Element):
    _children_name = 'versions'

    def __init__(self, name=None, versions=None, **kwargs):
        super().__init__(**kwargs)

        if name is None:
            name = getattr(self.__class__, 'name', '')
        self.name = name

        self._initialize_children(locals())

    def accept(self, visitor):
        visitor.visit_api(self)
        for version in self.versions:
            version.accept(visitor)



from .base import Element


class Field(Element):
    # noinspection PyShadowingBuiltins
    def __init__(self, name=None, type=None, required=None, **kwargs):
        super().__init__(**kwargs)
        if required is None:
            required = getattr(self, 'required', False)
        self.required = required

        if name is None:
            name = getattr(self, 'name', '')
        self.name = name

        if type is None:
            type = getattr(self, 'type', self.__class__)
        self.type = type

    def accept(self, visitor):
        visitor.visit_field(self)


class IntegerField(Field):
    def __init__(self, min_value=None, max_value=None, **kwargs):
        super().__init__(**kwargs)

        if min_value is None:
            min_value = getattr(self, 'min_value', None)
        self.min_value = min_value

        if max_value is None:
            max_value = getattr(self, 'max_value', None)
        self.max_value = max_value


class PositiveIntegerField(IntegerField):
    min_value = 0


class DecimalField(Field):
    def __init__(self, precision=None, **kwargs):
        super().__init__(**kwargs)

        if precision is None:
            try:
                precision = getattr(self, 'precision')
            except AttributeError:
                raise TypeError('Missing precision specification.')

        self.precision = precision


class StringField(Field):
    pass


class TimestampField(Field):
    pass


class ResourceField(Field):
    def __init__(self, name=None, type=None, resource=None):
        super().__init__(name, type)
        if resource is None:
            try:
                resource = getattr(self, 'resource')
            except AttributeError:
                raise TypeError('Missing resource specification.')
        self.resource = resource

    def accept(self, visitor):
        visitor.visit_field(self)
        self.resource.accept(visitor)



import pytest

from pactum.fields import Field
from pactum.resources import ListResource, Resource


def test_basic_resource():
    resource = Resource(
        name="TestResource",
        fields=[],
        behaviors=[],
    )

    assert resource.name == "TestResource"
    assert len(resource.fields) == 0
    assert len(resource.behaviors) == 0


def test_basic_resource_class_definition():
    class TestResource(Resource):
        name = "TestResource"
        fields = []
        behaviors = []

    resource = TestResource()
    assert resource.name == "TestResource"
    assert len(resource.fields) == 0
    assert len(resource.behaviors) == 0


def test_basic_resource_class_definition_with_class_name():
    class MyResource(Resource):
        fields = []
        behaviors = []

    resource = MyResource()
    assert resource.name == "MyResource"
    assert len(resource.fields) == 0
    assert len(resource.behaviors) == 0


def test_access_fields_by_key():
    class MyResource(Resource):
        fields = [
            Field(name='my_field'),
            Field(name='my_other_field')
        ]

    resource = MyResource()

    assert isinstance(resource['my_field'], Field)
    assert isinstance(resource['my_other_field'], Field)

    assert resource['my_field'].name == 'my_field'
    assert resource['my_other_field'].name == 'my_other_field'

    assert resource.fields[0].parent == resource
    assert resource.fields[1].parent == resource


def test_duplicate_field_names_raises_error():
    class MyResource(Resource):
        fields = [
            Field(name='my_field'),
            Field(name='my_field')
        ]

    with pytest.raises(AttributeError):
        MyResource()


def test_subclassing_add_fields():
    class MyResource(Resource):
        fields = [
            Field(name='my_field'),
        ]

    class MySubResource(MyResource):
        fields = MyResource.fields + [
            Field(name='my_other_field')
        ]

    resource = MySubResource()

    assert isinstance(resource['my_field'], Field)
    assert resource['my_field'].name == 'my_field'

    assert isinstance(resource['my_other_field'], Field)
    assert resource['my_other_field'].name == 'my_other_field'


def test_basic_list_resource(resource):
    list_resource = ListResource(resource=resource)

    assert list_resource.name == "ListResource"
    assert list_resource.resource == resource

    assert hasattr(list_resource, 'fields') is False
    assert hasattr(list_resource, 'behaviors') is False


def test_basic_list_resource_class_definition(resource):
    base_resource = resource

    class MyListResource(ListResource):
        name = "ListResource"
        resource = base_resource

    list_resource = MyListResource()

    assert list_resource.name == "ListResource"
    assert list_resource.resource == resource

    assert hasattr(list_resource, 'fields') is False
    assert hasattr(list_resource, 'behaviors') is False


def test_fail_list_resource_with_no_resource():
    with pytest.raises(TypeError):
        ListResource(name="ListResource")



class BaseResource:
    def __init__(self, name=None):
        if name is None:
            name = getattr(self, "name", self.__class__.__name__)
        self.name = name


class Resource(BaseResource):
    def __init__(self, name=None, fields=None, behaviors=None):
        super().__init__(name)
        self._mapfields = {}

        if fields is None:
            fields = getattr(self.__class__, "fields", [])

        for field in fields:
            if self._mapfields.get(field.name):
                raise AttributeError("Duplicate field names")
            self._mapfields[field.name] = field
            field.parent = self
        self.fields = fields

        if behaviors is None:
            behaviors = getattr(self, "behavior", [])
        self.behaviors = behaviors

    def __getitem__(self, item):
        return self._mapfields[item]

    def accept(self, visitor):
        visitor.visit_resource(self)
        for field in self.fields:
            field.accept(visitor)


class ListResource(BaseResource):
    def __init__(self, resource=None, **kwargs):
        super().__init__(**kwargs)

        if resource is None:
            try:
                resource = getattr(self, "resource")
            except AttributeError:
                raise TypeError("Missing resource specification.")

        self.resource = resource

    def accept(self, visitor):
        visitor.visit_list_resource(self)
        self.resource.accept(visitor)



import pytest

from pactum.route import Route


def test_basic_route():
    route = Route('/test/', actions=[])

    assert route.path == '/test/'
    assert len(route.actions) == 0


def test_route_class_definition():
    class TestRoute(Route):
        path = '/test/'
        actions = []

    route = TestRoute()

    assert route.path == '/test/'
    assert len(route.actions) == 0


def test_prefer_parameter_to_class_definition(action):
    class TestRoute(Route):
        path = '/test/'
        actions = []

    route = TestRoute(
        path="/test_by_param/",
        actions=[action]
    )

    assert len(route.actions) == 1
    assert route.path == "/test_by_param/"
    assert route.actions[0].parent == route


def test_fail_route_with_no_path(resource):
    with pytest.raises(TypeError):
        Route(actions=[])



#!/usr/bin/env python


import codecs
import os
import sys

from setuptools import find_packages, setup, Command
from shutil import rmtree

NAME = 'pactum'
VERSION = '0.1.1'
DESCRIPTION = 'Create API specifications and documentation using Python'
URL = 'https://github.com/olist/pactum'
EMAIL = 'pactum@olist.com'
AUTHOR = 'Olist Developers'
REQUIRED = ['Click', 'PyYAML']

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()


class PublishCommand(Command):
    """Support setup.py publish."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.sep.join(('.', 'dist')))
        except FileNotFoundError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal '.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine…')
        os.system('twine upload dist/*')

        sys.exit()


setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    packages=find_packages(),
    install_requires=REQUIRED,
    include_package_data=True,
    license='GPLv3',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Testing',
    ],
    cmdclass={
        'publish': PublishCommand,
    },
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    py_modules=['commands.openapi_export'],
    entry_points='''
        [console_scripts]
        pactum-openapi=commands.openapi_export:openapi_export
    '''
)



PATCH = 'PATCH'  # RFC 5789

CONNECT = 'CONNECT'  # RFC 2616
DELETE = 'DELETE'  # RFC 2616
GET = 'GET'  # RFC 2616
OPTIONS = 'OPTIONS'  # RFC 2616
HEAD = 'HEAD'  # RFC 2616
POST = 'POST'  # RFC 2616
PUT = 'PUT'  # RFC 2616
TRACE = 'TRACE'  # RFC 2616


COPY = 'COPY'  # RFC 2518 (WebDAV)
LOCK = 'LOCK'  # RFC 2518 (WebDAV)
MKCOL = 'MKCOL'  # RFC 2518 (WebDAV)
MOVE = 'MOVE'  # RFC 2518 (WebDAV)
PROPFIND = 'PROPFIND'  # RFC 2518 (WebDAV)
PROPPATCH = 'PROPPATCH'  # RFC 2518 (WebDAV)
UNLOCK = 'UNLOCK'  # RFC 2518 (WebDAV)

VERSION_CONTROL = 'VERSION-CONTROL'  # RFC 3253 (WebDAV Extension)
REPORT = 'REPORT'  # RFC 3253 (WebDAV Extension)
CHECKOUT = 'CHECKOUT'  # RFC 3253 (WebDAV Extension)
CHECKIN = 'CHECKIN'  # RFC 3253 (WebDAV Extension)
UNCHECKOUT = 'UNCHECKOUT'  # RFC 3253 (WebDAV Extension)
MKWORKSPACE = 'MKWORKSPACE'  # RFC 3253 (WebDAV Extension)
UPDATE = 'UPDATE'  # RFC 3253 (WebDAV Extension)
LABEL = 'LABEL'  # RFC 3253 (WebDAV Extension)
MERGE = 'MERGE'  # RFC 3253 (WebDAV Extension)
BASELINE_CONTROL = 'BASELINE-CONTROL'  # RFC 3253 (WebDAV Extension)
MKACTIVITY = 'MKACTIVITY'  # RFC 3253 (WebDAV Extension)

ORDERPATCH = 'ORDERPATCH'  # RFC 3648 (WebDAV Order Collection Spaecification)

ACL = 'ACL'  # RFC 3744 (WebDAV Access Control)



import pytest

from pactum import verbs
from pactum.action import Action
from pactum.route import Route
from pactum.request import Request
from pactum.version import Version


def test_basic_version_with_routes():
    version = Version(
        name="1.0",
        routes=[],
    )

    assert version.name == "1.0"
    assert len(version.routes) == 0


def test_version_class_definition_with_routes():
    class V1(Version):
        name = "V1"
        routes = []

    version = V1()

    assert version.name == "V1"
    assert len(version.routes) == 0


def test_fail_if_no_routes_and_no_version():
    with pytest.raises(TypeError):
        Version(name="v1")


def test_class_def_fails_if_no_routes():
    class V1(Version):
        name = "V1"

    with pytest.raises(TypeError):
        V1()


def test_prefer_parameter_to_class_definition(route):
    class TestVersion(Version):
        name = "Test Version"
        routes = []

    version = TestVersion(
        name="Test Version by parameter",
        routes=[route]
    )

    assert len(version.routes) == 1
    assert version.name == "Test Version by parameter"
    assert version.routes[0].parent == version


def test_validate_ambiguous_routes_on_version_init(resource):
    action1 = Action(Request(verb=verbs.GET), responses=[])
    action2 = Action(Request(verb=verbs.POST), responses=[])
    action3 = Action(Request(verb=verbs.GET), responses=[])
    route1 = Route('/route/', actions=[action1, action2])
    route2 = Route('/route/', actions=[action3])

    class TestVersion(Version):
        name = "v1"
        routes = [route1, route2]

    with pytest.raises(AttributeError):
        TestVersion()



from .import fields, verbs  # NOQA
from .action import Action  # NOQA
from .api import API  # NOQA
from .resources import ListResource, Resource  # NOQA
from .response import Response  # NOQA
from .request import Request  # NOQA
from .route import Route  # NOQA
from .version import Version  # NOQA



from .base import Element


class Version(Element):
    _children_name = 'routes'

    def __init__(self, name=None, routes=None):
        if name is None:
            name = getattr(self, 'name', '')
        self.name = name

        if routes is None:
            try:
                routes = getattr(self.__class__, 'routes')
            except AttributeError:
                raise TypeError("Version must have routes.")

        actions = set()
        for route in routes:
            for action in route.actions:
                action = (route.path, action.request.verb)
                if action in actions:
                    raise AttributeError('Ambiguous route and action request specification.')
                actions.add(action)

        self._initialize_children(locals())

    def accept(self, visitor):
        visitor.visit_version(self)
        for route in self.routes:
            route.accept(visitor)
